/* $Id: milter-checkrcpt.c,v 1.2 2011/07/12 06:15:35 dhartmei Exp $ */

/*
 * Copyright (c) 2007-2011 Daniel Hartmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

static const char rcsid[] = "$Id: milter-checkrcpt.c,v 1.2 2011/07/12 06:15:35 dhartmei Exp $";

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <libmilter/mfapi.h>

extern void	 die(const char *);

static int		 debug = 0;
static char		 server[32];
static char		 helo[256];
static char		 mailfrom[256];

#define	MAX_RCPT	128

struct context {
	char		 host_name[128];
	char		 host_addr[64];
	char		 helo[128];
	char		 hdr_from[128];
	char		 hdr_to[128];
	char		 hdr_subject[128];
	char		*rcpt_valid[MAX_RCPT];
	char		*rcpt_invalid[MAX_RCPT];
};

static sfsistat		 cb_connect(SMFICTX *, char *, _SOCK_ADDR *);
static sfsistat		 cb_helo(SMFICTX *, char *);
static sfsistat		 cb_envfrom(SMFICTX *, char **);
static sfsistat		 cb_envrcpt(SMFICTX *, char **);
static sfsistat		 cb_header(SMFICTX *, char *, char *);
static sfsistat		 cb_eom(SMFICTX *);
static sfsistat		 cb_close(SMFICTX *);
static void		 usage(const char *);
static void		 msg(int, struct context *, const char *, ...);

#define USER		"_milter-checkrcpt"
#define OCONN		"unix:/var/spool/milter-checkrcpt/sock"
#define RCODE_REJECT	"554"
#define RCODE_TEMPFAIL	"451"
#define XCODE_REJECT	"5.7.1"
#define XCODE_TEMPFAIL	"4.7.1"
#define	MAXRS		16

#define	ST_MTIME st_mtimespec

static int
check_rcpt(struct context *context, const char *rcpt)
{
	int state = 0;
	int fd;
	struct sockaddr_in sa;
	time_t timeout = time(NULL) + 10;
	char line[8192];
	int off = 0;

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		msg(LOG_ERR, context, "check_rcpt: socket: %s",
		    strerror(errno));
		return (0);
	}
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(server);
	sa.sin_port = htons(25);
	msg(LOG_DEBUG, context, "check_rcpt: connecting to %s", server);
	if (connect(fd, (struct sockaddr *)&sa, sizeof(sa))) {
		msg(LOG_ERR, context, "check_rcpt: connect: %s",
		    strerror(errno));
		close(fd);
		return (0);
	}
	if (fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK)) {
		msg(LOG_ERR, context, "check_rcpt: fcntl: %s",
		    strerror(errno));
		close(fd);
		return (0);
	}
	while (state < 9 && time(NULL) < timeout) {
		fd_set readfds, writefds;
		struct timeval tv;
		int r;

		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		FD_ZERO(&writefds);
		FD_SET(fd, &writefds);
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		r = select(fd + 1, &readfds, &writefds, NULL, &tv);
		if (r < 0) {
			if (errno != EINTR) {
				msg(LOG_ERR, context, "check_rcpt: select: %s",
				    strerror(errno));
				break;
			}
			usleep(10000);
			continue;
		} else if (r == 0)
			continue;
		/* r > 0 */
		if (FD_ISSET(fd, &readfds)) {
			char buf[8192];
			int len, i;

			len = read(fd, buf, sizeof(buf));
			if (len < 0) {
				if (errno != EINTR) {
					msg(LOG_ERR, context, "check_rcpt: "
					    "read: %s", strerror(errno));
					break;
				}
			} else if (len == 0) {
				msg(LOG_ERR, context, "check_rcpt: "
				    "connection reset");
				break;
			}
			for (i = 0; i < len; ++i) {
				if (buf[i] == '\n' || off >= sizeof(line) - 1) {
					line[off] = 0;
					msg(LOG_DEBUG, context, "check_rcpt: "
					    "read: %s", line);
					line[3] = 0;
					r = atoi(line);
					if (state % 2 == 0) {
						if (r >= 200 && r <= 299)
							state++;
						else
							state = 10;
					}
					off = 0;
				} else
					line[off++] = buf[i];
			}
		}
		if (FD_ISSET(fd, &writefds)) {
			char buf[8192];
			int off = 0, len;

			switch (state) {
			case 1:
				snprintf(buf, sizeof(buf), "HELO %s\r\n", helo);
				break;
			case 3:
				snprintf(buf, sizeof(buf), "MAIL FROM:<%s>\r\n",				    mailfrom);
				break;
			case 5:
				snprintf(buf, sizeof(buf), "RCPT TO:%s\r\n",
				    rcpt);
				break;
			case 7:
				strlcpy(buf, "QUIT\r\n", sizeof(buf));
				break;
			default:
				buf[0] = 0;
			}
			if (!buf[0]) {
				usleep(10000);
				continue;
			}
			while (off < strlen(buf)) {
				len = write(fd, buf + off, strlen(buf) - off);
				if (len < 0) {
					if (errno != EINTR) {
						msg(LOG_ERR, context,
						    "check_rcpt: write: %s",
						    strerror(errno));
						break;
					}
				}
				off += len;
			}
			if (off < strlen(buf)) {
				msg(LOG_ERR, context,
				    "check_rcpt: short write");
				break;
			}
			msg(LOG_DEBUG, context, "check_rcpt: wrote: %s", buf);
			state++;
		}
	}

	close(fd);
	if (state == 9) {
		msg(LOG_DEBUG, context, "check_rcpt('%s') OK", rcpt);
		return (0);
	} else {
		msg(LOG_DEBUG, context, "check_rcpt('%s') FAIL (state %d)",
		    rcpt, state);
		return (1);
	}
}

static sfsistat
cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
{
	struct context *context;

	context = calloc(1, sizeof(*context));
	if (context == NULL) {
		msg(LOG_ERR, NULL, "cb_connect: calloc: %s", strerror(errno));
		return (SMFIS_ACCEPT);
	}
	if (smfi_setpriv(ctx, context) != MI_SUCCESS) {
		free(context);
		msg(LOG_ERR, NULL, "cb_connect: smfi_setpriv");
		return (SMFIS_ACCEPT);
	}
	strlcpy(context->host_name, name, sizeof(context->host_name));
	strlcpy(context->host_addr, "unknown", sizeof(context->host_addr));
	if (sa) {
		switch (sa->sa_family) {
		case AF_INET: {
			struct sockaddr_in *sin = (struct sockaddr_in *)sa;

			if (inet_ntop(AF_INET, &sin->sin_addr.s_addr,
			    context->host_addr, sizeof(context->host_addr)) ==
			    NULL)
				msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
				    strerror(errno));
			break;
		}
		case AF_INET6: {
			struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;

			if (inet_ntop(AF_INET6, &sin6->sin6_addr,
			    context->host_addr, sizeof(context->host_addr)) ==
			    NULL)
				msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
				    strerror(errno));
			break;
		}
		}
	}
	if (!strncmp(context->host_addr, "10.", 3) ||
	    !strcmp(context->host_addr, "127.0.0.1") ||
	    !strcmp(context->host_addr, "::1"))
		return (SMFIS_ACCEPT);
	msg(LOG_DEBUG, context, "cb_connect('%s', '%s')",
	    context->host_name, context->host_addr);
	return (SMFIS_CONTINUE);
}

static sfsistat
cb_helo(SMFICTX *ctx, char *arg)
{
	struct context *context;
	int i;

	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
		msg(LOG_ERR, NULL, "cb_helo: smfi_getpriv");
		return (SMFIS_ACCEPT);
	}
	strlcpy(context->helo, arg, sizeof(context->helo));
	msg(LOG_DEBUG, context, "cb_helo('%s')", arg);
	/* multiple HELO imply RSET in sendmail */
	for (i = 0; i < MAX_RCPT; ++i)
		context->rcpt_valid[i] = context->rcpt_invalid[i] = NULL;
	return (SMFIS_CONTINUE);
}

static sfsistat
cb_envfrom(SMFICTX *ctx, char **args)
{
	struct context *context;
	int i;

	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
		msg(LOG_ERR, NULL, "cb_envfrom: smfi_getpriv");
		return (SMFIS_ACCEPT);
	}
	/* multiple MAIL FROM indicate separate messages */
	for (i = 0; i < MAX_RCPT; ++i)
		context->rcpt_valid[i] = context->rcpt_invalid[i] = NULL;
	if (*args != NULL)
		msg(LOG_DEBUG, context, "cb_envfrom('%s')", *args);
	return (SMFIS_CONTINUE);
}

static sfsistat
cb_envrcpt(SMFICTX *ctx, char **args)
{
	struct context *context;
	int i;

	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
		msg(LOG_ERR, NULL, "cb_envrcpt: smfi_getpriv");
		return (SMFIS_ACCEPT);
	}
	/* multiple RCPT TO: possible */
	for (; *args != NULL; ++args) {
		msg(LOG_DEBUG, context, "cb_envrcpt('%s')", *args);
		if (check_rcpt(context, *args)) {
			msg(LOG_DEBUG, context, "Invalid rcpt: %s", *args);
			for (i = 0; i < MAX_RCPT; ++i)
				if (context->rcpt_invalid[i] == NULL)
					break;
			if (i < MAX_RCPT)
				context->rcpt_invalid[i] = strdup(*args);
		} else {
			msg(LOG_DEBUG, context, "Valid rcpt: %s", *args);
			for (i = 0; i < MAX_RCPT; ++i)
				if (context->rcpt_valid[i] == NULL)
					break;
			if (i < MAX_RCPT)
				context->rcpt_valid[i] = strdup(*args);
		}
	}
	return (SMFIS_CONTINUE);
}

static sfsistat
cb_header(SMFICTX *ctx, char *name, char *value)
{
	struct context *context;

	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
		msg(LOG_ERR, context, "cb_header: smfi_getpriv");
		return (SMFIS_ACCEPT);
	}
	msg(LOG_DEBUG, context, "cb_header('%s', '%s')", name, value);
	if (!strcasecmp(name, "From"))
		strlcpy(context->hdr_from, value, sizeof(context->hdr_from));
	else if (!strcasecmp(name, "To"))
		strlcpy(context->hdr_to, value, sizeof(context->hdr_to));
	else if (!strcasecmp(name, "Subject"))
		strlcpy(context->hdr_subject, value,
		    sizeof(context->hdr_subject));
	return (SMFIS_CONTINUE);
}

static sfsistat
cb_eom(SMFICTX *ctx)
{
	struct context *context;
	int i;

	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
		msg(LOG_ERR, NULL, "cb_eom: smfi_getpriv");
		return (SMFIS_ACCEPT);
	}
	msg(LOG_DEBUG, context, "cb_eom()");
	if (context->rcpt_valid[0] == NULL) {
		msg(LOG_NOTICE, context, "REJECT, Helo: %s, Rcpt: %s, "
		    "From: %s, To: %s, Subject: %s", context->helo,
		    context->rcpt_invalid[0] == NULL ? "none" :
		    context->rcpt_invalid[0], context->hdr_from,
		    context->hdr_to, context->hdr_subject);
		if (smfi_setreply(ctx, RCODE_REJECT, XCODE_REJECT,
		    "No valid recipients") != MI_SUCCESS)
			msg(LOG_ERR, context, "smfi_setreply");
		return (SMFIS_REJECT);
	}
	msg(LOG_INFO, context, "ACCEPT, Helo: %s, Rcpt: %s, From: %s, "
	    "To: %s, Subject: %s", context->helo,
	    context->rcpt_valid[0] == NULL ? "none" :
	    context->rcpt_valid[0], context->hdr_from, context->hdr_to,
	    context->hdr_subject);
	for (i = 0; i < MAX_RCPT && context->rcpt_invalid[i] != NULL; ++i) {
		msg(LOG_INFO, context, "DELRCPT, Rcpt: %s",
		    context->rcpt_invalid[i]);
		if (smfi_delrcpt(ctx, context->rcpt_invalid[i]) != MI_SUCCESS)
			msg(LOG_ERR, context, "smfi_delrcpt");
	}
	return (SMFIS_ACCEPT);
}

static sfsistat
cb_close(SMFICTX *ctx)
{
	struct context *context;

	context = (struct context *)smfi_getpriv(ctx);
	msg(LOG_DEBUG, context, "cb_close()");
	if (context != NULL) {
		int i;

		smfi_setpriv(ctx, NULL);
		for (i = 0; i < MAX_RCPT; ++i) {
			if (context->rcpt_valid[i] != NULL)
				free(context->rcpt_valid[i]);
			if (context->rcpt_invalid[i] != NULL)
				free(context->rcpt_invalid[i]);
		}
		free(context);
	}
	return (SMFIS_CONTINUE);
}

struct smfiDesc smfilter = {
	"milter-checkrcpt",	/* filter name */
	SMFI_VERSION,		/* version code -- do not change */
	SMFIF_DELRCPT,	 	/* flags */
	cb_connect,		/* connection info filter */
	cb_helo,		/* SMTP HELO command filter */
	cb_envfrom,		/* envelope sender filter */
	cb_envrcpt,		/* envelope recipient filter */
	cb_header,		/* header filter */
	NULL,			/* end of header */
	NULL,			/* body block */
	cb_eom,			/* end of message */
	NULL,			/* message aborted */
	cb_close		/* connection cleanup */
};

static void
msg(int priority, struct context *context, const char *fmt, ...)
{
	va_list ap;
	char msg[8192];

	va_start(ap, fmt);
	if (context != NULL)
		snprintf(msg, sizeof(msg), "%s: ", context->host_addr);
	else
		msg[0] = 0;
	vsnprintf(msg + strlen(msg), sizeof(msg) - strlen(msg), fmt, ap);
	if (debug)
		printf("syslog: %s\n", msg);
	else
		syslog(priority, "%s", msg);
	va_end(ap);
}

static void
usage(const char *argv0)
{
	fprintf(stderr, "usage: %s [-d] [-h helo] [-m mailfrom] [-p pipe] [-r rcptto] [-u user] -s addr\n", argv0);
	exit(1);
}

void
die(const char *reason)
{
	msg(LOG_ERR, NULL, "die: %s", reason);
	smfi_stop();
	sleep(60);
	/* not reached, smfi_stop() kills thread */
	abort();
}

int
main(int argc, char **argv)
{
	int ch;
	const char *oconn = OCONN;
	const char *user = USER;
	sfsistat r = MI_FAILURE;
	const char *ofile = NULL;
	const char *rcptto = NULL;

	tzset();
	openlog("milter-checkrcpt", LOG_PID | LOG_NDELAY, LOG_DAEMON);

	while ((ch = getopt(argc, argv, "dh:m:p:r:s:u:")) != -1) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 'h':
			strlcpy(helo, optarg, sizeof(helo));
			break;
		case 'm':
			strlcpy(mailfrom, optarg, sizeof(mailfrom));
			break;
		case 'p':
			oconn = optarg;
			break;
		case 'r':
			rcptto = optarg;
			break;
		case 's':
			strlcpy(server, optarg, sizeof(server));
			break;
		case 'u':
			user = optarg;
			break;
		default:
			usage(argv[0]);
		}
	}
	if (argc != optind) {
		fprintf(stderr, "unknown command line argument: %s ...",
		    argv[optind]);
		usage(argv[0]);
	}
	if (!server[0])
		usage(argv[0]);
	if (!helo[0] && gethostname(helo, sizeof(helo))) {
		fprintf(stderr, "gethostname: %s\n", strerror(errno));
		return (1);
	}
	if (!mailfrom[0])
		snprintf(mailfrom, sizeof(mailfrom), "milter-checkrcpt@%s", helo);

	if (!strncmp(oconn, "unix:", 5))
		ofile = oconn + 5;
	else if (!strncmp(oconn, "local:", 6))
		ofile = oconn + 6;
	if (ofile != NULL)
		unlink(ofile);

	/* drop privileges */
	if (!getuid()) {
		struct passwd *pw;

		if ((pw = getpwnam(user)) == NULL) {
			fprintf(stderr, "getpwnam: %s: %s\n", user,
			    strerror(errno));
			return (1);
		}
		setgroups(1, &pw->pw_gid);
		if (setegid(pw->pw_gid) || setgid(pw->pw_gid)) {
			fprintf(stderr, "setgid: %s\n", strerror(errno));
			return (1);
		}
		if (
		    seteuid(pw->pw_uid) || setuid(pw->pw_uid)) {
			fprintf(stderr, "setuid: %s\n", strerror(errno));
			return (1);
		}
	}

	if (rcptto) {
		printf("checking recipient %s on server %s\n", rcptto, server);
		if (check_rcpt(NULL, rcptto)) {
			printf("FAIL\n");
			return (1);
		} else {
			printf("OK\n");
			return (0);
		}
	}

	if (smfi_setconn((char *)oconn) != MI_SUCCESS) {
		fprintf(stderr, "smfi_setconn: %s: failed\n", oconn);
		goto done;
	}

	if (smfi_register(smfilter) != MI_SUCCESS) {
		fprintf(stderr, "smfi_register: failed\n");
		goto done;
	}

	/* daemonize (detach from controlling terminal) */
	if (!debug && daemon(0, 0)) {
		fprintf(stderr, "daemon: %s\n", strerror(errno));
		goto done;
	}
	umask(0177);

	msg(LOG_INFO, NULL, "started: %s", rcsid);
	r = smfi_main();
	if (r != MI_SUCCESS)
		msg(LOG_ERR, NULL, "smfi_main: terminating due to error");
	else
		msg(LOG_INFO, NULL, "smfi_main: terminating without error");

done:
	return (r);
}
